"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bluebird_1 = __importDefault(require("bluebird"));
const TimeoutError_1 = require("./TimeoutError");
const constants_1 = require("./constants");
class Queue {
    constructor(options) {
        var _a;
        this.taskPointer = 0;
        this.isIdle = true;
        this.tasks = [];
        this.taskTimeout = options === null || options === void 0 ? void 0 : options.taskTimeout;
        this.gcThreshold = (_a = options === null || options === void 0 ? void 0 : options.gcThreshold) !== null && _a !== void 0 ? _a : constants_1.DEFAULT_GC_THRESHOLD;
    }
    setTaskFunction(input) {
        this.taskFunction = input;
        return this;
    }
    deleteTaskFunction() {
        delete this.taskFunction;
        return this;
    }
    getNextTask() {
        if (this.taskPointer > this.gcThreshold) {
            this.gc();
        }
        return this.tasks[this.taskPointer++];
    }
    hasTasks() {
        return this.tasks.length > this.taskPointer;
    }
    gc() {
        if (this.hasTasks()) {
            this.tasks = this.tasks.slice(this.taskPointer);
        }
        else {
            this.tasks = [];
        }
        this.taskPointer = 0;
        return this;
    }
    startLoop() {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            this.isIdle = false;
            while (this.hasTasks()) {
                const task = this.getNextTask();
                if (!task) {
                    continue;
                }
                const timeout = (_b = (_a = task.options) === null || _a === void 0 ? void 0 : _a.timeout) !== null && _b !== void 0 ? _b : this.taskTimeout;
                const taskFunction = (_c = task.taskFunction) !== null && _c !== void 0 ? _c : this.taskFunction;
                const onComplete = task.onComplete;
                if (!taskFunction) {
                    continue;
                }
                try {
                    let result = taskFunction(task.taskData);
                    if (timeout) {
                        result = bluebird_1.default.resolve(result).timeout(timeout, new TimeoutError_1.TimeoutError(`Task timeout [${timeout}ms]`));
                    }
                    else {
                        result = Promise.resolve(result);
                    }
                    if (onComplete) {
                        yield result.then(function (result) { onComplete(undefined, result); });
                    }
                    else {
                        yield result;
                    }
                }
                catch (err) {
                    onComplete && onComplete(err);
                }
                yield bluebird_1.default.resolve();
            }
            this.gc();
            this.isIdle = true;
        });
    }
    addTaskData(taskData, options) {
        if (!(options === null || options === void 0 ? void 0 : options.ignoreResult)) {
            return new bluebird_1.default((resolve, reject) => {
                this.tasks.push({
                    taskData,
                    options,
                    onComplete(err, result) {
                        err ? reject(err) : resolve(result);
                    }
                });
                this.isIdle && this.startLoop();
            });
        }
        this.tasks.push({ taskData, options });
        this.isIdle && this.startLoop();
    }
    addTask(taskFunction, options) {
        if (!(options === null || options === void 0 ? void 0 : options.ignoreResult)) {
            return new bluebird_1.default((resolve, reject) => {
                this.tasks.push({
                    taskFunction,
                    options,
                    onComplete(err, result) {
                        err ? reject(err) : resolve(result);
                    }
                });
                this.isIdle && this.startLoop();
            });
        }
        this.tasks.push({ taskFunction, options });
        this.isIdle && this.startLoop();
    }
}
exports.Queue = Queue;
